using Microsoft.AspNetCore.Authentication.Cookies;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using System;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication.OpenIdConnect;
using Microsoft.IdentityModel.Protocols.OpenIdConnect;
using Newtonsoft.Json;
using System.Collections.Generic;
using System.Net.Http;
using Microsoft.AspNetCore.Authentication;
using System.Web;

namespace LibrarianWebApp
{
   public class Startup
   {
      public Startup(IConfiguration configuration)
      {
         Configuration = configuration;
      }

      public IConfiguration Configuration { get; }
      public static string ClaimsIssuer { get; set; } = string.Empty;
      public static string EndSessionEndpoint { get; set; } = string.Empty;
      public static string AuthorizeEndpoint { get; set; } = string.Empty;
      public static string TokenEndpoint { get; set; } = string.Empty;

      // This method gets called by the runtime. Use this method to add services to the container.
      public void ConfigureServices(IServiceCollection services)
      {
         string barTenderCloudCluster = Configuration["BarTenderCloudCluster"];

         RetrieveAuthenticationConfiguration(barTenderCloudCluster);

         // Add authentication services.
         services.AddAuthentication(options =>
         {
            options.DefaultAuthenticateScheme = CookieAuthenticationDefaults.AuthenticationScheme;
            options.DefaultSignInScheme = CookieAuthenticationDefaults.AuthenticationScheme;
            options.DefaultChallengeScheme = CookieAuthenticationDefaults.AuthenticationScheme;
         })
         .AddCookie()
         .AddOpenIdConnect("BarTenderCloud", options =>
         {
            // Set the authority to your identity provider domain.
            options.Authority = ClaimsIssuer;

            // Configure the identity provider client ID and client secret.
            options.ClientId = Configuration["ClientId"];
            options.ClientSecret = Configuration["ClientSecret"];

            // Set the response type to "code."
            options.ResponseType = OpenIdConnectResponseType.Code;

            // Configure the scope.
            options.Scope.Clear();
            options.Scope.Add("openid");
            options.Scope.Add("BarTenderServiceApi");

            // Set the callback path, so that the identity provider will call back to http://localhost:44300/callback.
            options.CallbackPath = new PathString("/callback");

            // Configure the claims issuer to be the identity provider.
            options.ClaimsIssuer = ClaimsIssuer;

            // Save tokens to AuthenticationProperties (required so that access_token is a JWT).
            options.SaveTokens = true;

            options.Events = new OpenIdConnectEvents
            {
               // Handle the logout redirection. 
               OnRedirectToIdentityProviderForSignOut = (context) =>
               {
                  string logoutUri = string.Empty;
                  string clientIdInfo = $"?client_id={options.ClientId}";
                  string redirectToQueryParam = string.Empty;
                  string redirectToPath = string.Empty;

                  if (!string.IsNullOrEmpty(EndSessionEndpoint))
                  {
                     logoutUri = EndSessionEndpoint;
                     redirectToQueryParam = "post_logout_redirect_uri";
                  }
                  else
                  {
                     // No end_session_endpoint and not an auth0 domain, unsupported Idp
                     throw new Exception("idp does not publish end_session_endpoint to support logout");
                  }

                  logoutUri += clientIdInfo;

                  var postLogoutUri = context.Properties.RedirectUri;
                  if (!string.IsNullOrEmpty(postLogoutUri))
                  {
                     if (postLogoutUri.StartsWith("/"))
                     {
                        // transform to absolute
                        var request = context.Request;
                        postLogoutUri = request.Scheme + "://" + request.Host + request.PathBase +
                                             postLogoutUri;
                     }

                     if (postLogoutUri.EndsWith("/"))
                        postLogoutUri = postLogoutUri.TrimEnd('/');

                     logoutUri += $"&{redirectToQueryParam}={HttpUtility.UrlEncode(postLogoutUri)}";
                  }

                  var idtok = context.HttpContext.GetTokenAsync("id_token").Result;
                  if (!string.IsNullOrEmpty(idtok))
                     logoutUri += $"&id_token_hint={idtok}";

                  context.Response.Redirect(logoutUri);
                  context.HandleResponse();

                  return Task.CompletedTask;
               },

               // Sign-in
               // (Required so that access_token is a JWT.)
               OnRedirectToIdentityProvider = context =>
               {
                  // The context's ProtocolMessage can be used to pass along additional query parameters
                  // to the identity provider's /authorize endpoint.
                  // 
                  // Set the audience query parameter to the API identifier to ensure that the returned access tokens can be used
                  // to call protected endpoints on the corresponding API.
                  context.ProtocolMessage.SetParameter("audience", Configuration["Audience"]);

                  // To skip the organization prompt set the organization's DNS name the request header 
                  //context.ProtocolMessage.SetParameter("OrganizationDnsName", "myorg1234");

                  return Task.FromResult(0);
               }
            };
         });

         services.AddControllersWithViews();
         services.AddRazorPages();
      }

      // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
      public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
      {
         if (env.IsDevelopment())
         {
            app.UseDeveloperExceptionPage();
         }
         else
         {
            app.UseExceptionHandler("/Librarian/Error");
            // The default HSTS value is 30 days. You may want to change this for production scenarios, see https://aka.ms/aspnetcore-hsts.
            app.UseHsts();
         }

         app.UseHttpsRedirection();
         app.UseStaticFiles();

         app.UseRouting();
         app.UseAuthentication();
         app.UseAuthorization();

         app.UseEndpoints(endpoints => { endpoints.MapDefaultControllerRoute(); endpoints.MapRazorPages(); });
      }

      void RetrieveAuthenticationConfiguration(string barTenderCloudCluster)
      {
         HttpClient httpClient = new HttpClient();
         string url = barTenderCloudCluster + (!barTenderCloudCluster.EndsWith('/') ? "/" : "") + ".well-known/openid-configuration";
         HttpResponseMessage msg = httpClient.GetAsync(url).Result;
         if (!msg.IsSuccessStatusCode)
            throw new Exception("Failure to retrieve OIDC discovery document");

         string result = msg.Content.ReadAsStringAsync().Result;
         IDictionary<string, object> dictionary = JsonConvert.DeserializeObject<IDictionary<string, object>>(result)!;
         if (dictionary.ContainsKey("issuer"))
         {
            var i = (string)dictionary["issuer"];
            ClaimsIssuer = i.EndsWith('/') ? i.TrimEnd(new char[] { '/' }) : i;
         }

         if (dictionary.ContainsKey("authorization_endpoint"))
         {
            var i = (string)dictionary["authorization_endpoint"];
            AuthorizeEndpoint = i.EndsWith('/') ? i.TrimEnd(new char[] { '/' }) : i;
         }

         if (dictionary.ContainsKey("end_session_endpoint"))
         {
            var i = (string)dictionary["end_session_endpoint"];
            EndSessionEndpoint = i.EndsWith('/') ? i.TrimEnd(new char[] { '/' }) : i;
         }

         if (dictionary.ContainsKey("token_endpoint"))
         {
            var i = (string)dictionary["token_endpoint"];
            TokenEndpoint = i.EndsWith('/') ? i.TrimEnd(new char[] { '/' }) : i;
         }
      }
   }
}
